/*	This file contains type definitions for generic and global structures,
and declarations of cetain global variables. The file is intended to be
included in many other files, as well as being compiled and loaded itself.
	For variables, arrays etc defined herein, the instantiation is
suppressed when included in other files by #define NOTGLOB 1 in those other
files. The declarations herein then become converted to "extern" declarations.
	*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#define Sf double
#define Sw int
#define Sh short
#define Sc char

#define Maxsamples 10  /* Max number of samples */
#define Maxvsets 3
#define Maxpoplns 15	/* Max number of popln models */
#define MaxClasses 150
#define MaxZ 100	/*  Length of the Zero vector  */
#define LL 300	/*	Length of input line buffer */

/*	-----------------  Control  ------------------------------   */

#define AdjTr 2
#define AdjPr 4
#define AdjAll 6
/*	The above define mask bits in 'control' to control
adjustment of weights, scores, structure, parameters
respectively  */


#define Partial 1
#define Most_likely 2
#define Random 3
/*	These are values for 'dfix' which normally
controls the mode of weight assignment  */



/*	-----------------  Variable types  ---------------------  */

typedef struct Vtypest {
	Sw id;
	Sw datsize;
	Sw vauxsize;	/* Size of aux block for vartype in vlist */
	Sw sauxsize;	/* size of aux block for vartype in sample */
	Sw pauxsize;	/* size of aux block for vartype in popln */
	Sc *name;
	Sw (*readvaux)();	/* Fun to read aux attribute info */
	Sw (*readsaux)();	/* Fun to read aux sample info */
	Sw (*readdat) ();	/* Fun to read a datum */
	void (*printdat) ();	/* Fun to print datum value */
	void (*setsizes) ();	/* To set basicsize, statssize */
	void (*clearstats) ();	/* To clear stats prior to re-estimation */
	void (*derivvar) ();
	void (*costvar) ();
	void (*adjust) ();
	void (*vprint) ();
	Sw (*inpar) ();		/* To read params from '.prep' type file */
	void (*setvar) ();
	} Vtype;


/*	-------------------  Files ----------------------------------  */

typedef struct Bufst	{
	FILE *cfile;
	Sw line, nch;
	Sc cname [80];
	Sc inl [LL];
	}	Buf;

/*	------------------  Allocation blocks  --------------------  */
typedef struct Blockst {
	struct Blockst *nblock;  Sw size;
	}  Block;


/*	-------------------  Attributes  -------------------------   */

typedef struct AVinstst {
	Sw id;
	Sw itype;
	Sw idle;	/* Inactive attribute flag */
	Sw basicsize;   /*  Sizeof basic block (CVinst) for this var */
	Sw statssize;   /* Sizeof stats block (EVinst) for this var */
	Vtype *vtp;
	char *vaux;
	Sc name [80];
	} AVinst;

typedef struct Vsetst {
	Sw id;
	Block *vblks;		/* Ptr to chain of blocks allocated */
	char dfname [80];  /* file name of vset */
	char name [80];
	Sw nv;	/* Number of variables */
	Sw nactv;	/* Number of active variables */
	AVinst *vlist;
	}	Vset;


/*	--------------------  Samples  ---------------------------   */

typedef struct SVinstst {
	Sw id;
	Sw nval;
	char *saux;
	Sw offset;	/*  offset of (missing, value) in record  */
	}  SVinst;


/*	Sample data is packed into a block of 'records' addressed by
the 'recs' pointer in a Sample structure. There is one record per thing in the
sample. Each record has the following format:
	char active_flag. If zero, the thing is ignored in building classes.
	Sw ident	The thing identifier, as a positive integer.
  Then follow 'nv' fields for the attribute values of the thing. Each field
actually has two parts:
	char missing_flag  If non-zero, shows value is unknown .
	Datum value.  The type Datum depends on the type of attribute. This
		field is present even if the missing flag is on, but contains
		garbage.
	*/

typedef struct Samplest	{
	Sw id;
	Block *sblks;	/* Ptr to chain of blocks allocated for sample */
	char vstname [80];	/* Name of variable-set */
	Sw nc;		/* Num of cases */
	Sw nact;	/* Num of active cases */
	SVinst *svars;	/* Ptr to vector of SVinsts, one per variable */
	char *recs;	/*  vector of records  */
	Sw reclen;	/*  Length in chars of a data record  */
	Sf bestcost;	/*  Cost of best model */
	Sw goodtime;	/*  Popln age when bestcost reached  */
	Sc name[80];
	Sc dfname[80];	/*  Data file name  */
	}  Sample;






/*	----------------------   Classes  ---------------------------     */


typedef struct CVinstst{ /*Structure for basic info on a var in a class*/
	Sw id;
	Sw signif;
	Sf samplesize;
	}	CVinst;


typedef struct EVinstst	{/* Stuff for var in class in expln */
	Sf cnt;		/*  Num of values */
	Sf stcost;
	Sf spcost;
	}	EVinst;



typedef struct Classst	{
	Sf relab;
	Sf mlogab;	/*  - log relab  */
	Sf cbcost;	/* Best total class cost  */
	Sf cbpcost;
	Sf cbtcost;
	Sf cdtcost;	/* cbtcost after adjustment */
	Sf cnt;		/* sum of weights of members  */
	Sc type;	/*  1 = popln, 8 = combine, 2 = leaf, 4 = sub */
	Sw age;		/*  age in massage counts  */
	Sw ison1, ison2;	/*  id's of subs  */
	Sw idad;	/* If a Sub, id of main else id of root */
	Sw serial;
	Sf newcnt;	/*  Accumulates weights for cnt  */
	Sf casecost;	/*  tcost of current case  */
	Sf casewt;	/*  weight of current case  */
/*	**************************
	Items below this line are set up when class is made by makeclass()
and should NOT be copied to a new class structure. IT IS ASSUMED THAT
'ID' IS THE FIRST ITEM BELOW THE LINE.
	********************* */
	Sw id;
	CVinst **basics;	/* ptr to vec of ptrs to variable basics */
	EVinst **stats;		/* ptr to vec of ptrs to variable stats blocks*/
	}	Class;

#define Pop 1
#define Leaf 2
#define Sub 4
#define TC 8
#define Vacant 64


/*	--------------------------  Population  ----------------------  */

typedef struct PVinstst	{
	Sw id;
	char *paux;
	}	PVinst;


typedef struct Poplnst	{
	Sw id;
	Block *pblks, *jblks;	/* Ptrs to bocks allocated for popln,
		and for popln as model of sample */
	char vstname [80];	/* Name of variable-set */
	Sc samplename[80];  /*  Name of sample to which popln is attached if any*/
	Sw nc;		/*  Size of sample attached, or 0 */
	Sw samplesize;	/*  num of active cases in sample used for training */
	Class **classes;	/* ptr to vec of ptrs to classes  */
	PVinst *pvars;	/* Ptr to vector of PVinsts, one per variable */
	Sc dfname[80];	/*  Popln file name  */
	Sc name[80];
	Sw nextserial;	/*  Next serial number for a new class */
	Sw ncl;		/* Number of classes  */
	Sw nleaf;	/* Number of leaves */
	Sw root;	/* index of popln class  */
	Sw tcic;     /* index of TC class or -1 if none */
	Sw mncl;	/*  Length of 'classes' vec. */
	Sw hicl;	/*  Highest allocated entry in pop->classes */
	Sf cost;	/*  Cost of model on current sample  */
	}  Popln;

/*	-------------------------------------------------------------   */

typedef struct Contextst	{
	Vset *vset;
	Sample *sample;
	Popln *popln;
	Buf *buffer;
	}  Context;


/*	--------- Functions -------------------------------   */

/*	In LISTEN.c	*/
	Sw hark (Sc *lline);
/*		end listen.c		*/

/*	In INPUTS.c  */
#ifndef INPUTS
	extern Sw terminator;
	extern Buf cfilebuf, commsbuf;
#endif
	Sw bufopen();
	Sw newline();
	Sw readint (Sw *x, Sw cnl);
	Sw readdf (Sf *x, Sw cnl);
	Sw readalf (Sc *str, Sw cnl);
	Sw readch (Sw cnl);
	void swallow ();
	void bufclose ();
	void revert (Sw flag);
	void rep (Sw ch);
	void flp ();
/*		end inputs.c		*/

/*	In POPLNS.c   */
	void nextclass (Class **ptr);
	Sw makepop();
	Sw firstpop();
	void makesubs (Sw kk);
	void setpop();
	void killpop (Sw px);
	Sw copypop (Sw p1, Sc *newname);
	Sw savepop (Sw p1, Sc *newname);
	Sw rsave (Sc *nam);
	Sw adv (FILE *fl);
	Sw rprep (Sc *nam);
	Sw loadpop (Sw pp);
	void printtree();
	Sw bestpopid ();
	void trackbest(Sw verify);
	Sw pname2id (Sc *nam);
	void crosstab (Sw xid);
	Sw rmrep (char* memname);
/*		end poplns.c		*/


/*	In CLASSES.c	*/
	Sw s2id (Sw ss);
	Sw makeclass ();
	void cleartcosts (Class *ccl);
	void costvarall (Class *ccl);
	void derivvarall (Class *ccl);
	void adjustclass (Class *ccl);
	void killsons (Sw kk);
	void printclass (Sw kk, Sw full, FILE* fl);
	void setclass (Class *ccl);
	Sw splitleaf (Sw kk);
	void clobber (Sw numc);
	Sw setsig (Sf ben);
	void prvhead (Class *ccl, Sw iv, FILE *fl);
	Sw   invhead (Class *ccl, Sw iv, FILE *fl);
/*		end classes.c		*/


/*	In DOALL.c	*/
	Sw doall (Sw ncy);
	void findall (Sw typ);
	void tidy (Sw hit);
	Sw uran();
	Sw sran ();
	Sf fran();
	void docase (Sw cse, Sw derivs);

#define Dead -10
#define Deadsmall -11
#define Deadsing  -12
#define Deadorphan -13
#define Deadkilled -14
/*		end doall.c		*/


/*	In TUNE.c	*/
	void defaulttune();
/*		end tune.c	*/


/*	In TACTICS.c	*/
	void flatten();
	void ranclass (Sw nn);
	void trycombine ();
	Sw combine (Sw ison1, Sw ison2);
/*		end tactics.c		*/


/*	In BADMOVES.c	*/
#define BadSize 1013
	 void clearbadm ();
	 Sw testbadm (Sw code, Sw w1, Sw w2);
	 void setbadm (Sw code, Sw s1, Sw s2);
/*		end badmoves.c		*/


/*	In BLOCK.c	*/
	void *gtsp (Sw gr, Sw size);
	void freesp (Sw gr);
	Sw repspace (Sw pp);
/*		end block.c		*/


/*	In DOTYPES.c	*/
	void dotypes ();
/*		end dotypes.c		*/


/*	In SAMPLES.c	*/
	void printdatum (Sw i, Sw n);
	Sw readvset ();
	Sw readsample (Sc *fname);
	Sw sname2id (Sc *nam, Sw expect);
	Sw vname2id (Sc *nam);
	Sw qssamp (Sample *samp);
	Sw id2ind (Sw id);
	Sw thinglist (Sc *tlstname);
	Sw memlist (Sc *mlstname);
	void sortbyser ();
/*		end samples.c		*/


/*	In GLOB.c	*/
	void cmcpy (void *pt, void *pf, Sw ntm);
 	Sc* sers (Class *cll);
/*		end glob.c		*/



/*	In MAIN.c     */
#ifndef MAIN
	extern void suffix (Sw jj, char* suff);
	extern char sparam[];
#endif


/*	In HTREE.c   */
	void **mmat (Sw row, Sw col, Sw isz);
	Sw htree ();


/*	-----------------  routines defined here  -------------------  */
#ifndef NOTGLOB

/*	A memory-to-memory copier replacing memcpy, which I can't use
properly on all systems.   */
void cmcpy (pt, pf, ntm)
	void *pt, *pf;  Sw ntm;
{
	char *cpt, *cpf;

	cpt = (char*) pt;  cpf = (char*) pf;
/*	Copy ntm chars from pf to pt  */
next:
	*cpt = *cpf;  cpt++;  cpf++;
	ntm--;  if (ntm) goto next;
	return;
}


/*	To assist in printing class serials  */
char decsers [8];

char* sers (cll)
	Class *cll;
{
	Sw i, j, k;
	for (i = 0; i < 6; i++) decsers[i] = ' ';  decsers[6] = 0;
	if (! cll) goto done;
	if (cll->type == Vacant) goto done;
	j = cll->serial;
	k = j & 3;
	if (k == 1) decsers[5] = 'a';
	else if (k == 2) decsers[5] = 'b';
	j = j >> 2;
	decsers[4] = '0';
	for (i = 4; i > 0; i--)	{
		if (j <= 0) goto done;
		decsers[i] = j % 10 + '0';  j = j / 10;
		}
	if (j) { strcpy (decsers+1, "2BIG"); }
done:	return (decsers);
}
#endif		/* end of function definitions in glob.c */


/*	--------------  Global variables declared here  ----------------  */
#ifdef NOTGLOB
#define Ex extern
#else
#define Ex
#endif

/*	mathematical constants   */
	Ex Sf hlg2pi, hlg2, pi, bit, bit2, twoonpi, pion2;

/*	general:	*/
	Ex Sw Ntypes;		/* The number of different attribute types */
	Ex Vtype *types;	/* a vector of Ntypes type definitions,
				created in dotypes */
	Ex Context ctx;	/* current context */
	Ex Vset *vsets[Maxvsets];
	Ex Sample *samples[Maxsamples];
	Ex Popln *poplns [Maxpoplns];
	Ex Sw nsamples;

/*	re inputs for main  */
	Ex Buf *source;		/* Ptr to command source buffer */

/*	re Sample records  */
	Ex char *rec;	/*  Common ptr to a data record  */
	Ex Sw thing;	/*  Index of current thing  */
	Ex char *loc;		/*  Common ptr to a data field  */
	Ex Sw reclen;	/*  reclen of current sample  */
	Ex char *recs;		/*  Common ptr to data records block of a sample */

/*	re hark  */
	Ex Sw heard;	/* Flag showing a new command line has been detected */
	Ex Sw usestdin;	/* Flag showing input from sdtdin */

	Ex Sw trapkk, trapage, trapcnt;

/*	To control what is adjusted  */
	Ex Sw control, dcontrol;

/*	To determine how weights are distributed  */
	Ex Sw dfix, fix;

/*	re Poplns  */
	Ex Vset *vst;
	Ex Sample *samp;
	Ex Popln *pop;
	Ex AVinst *avi;
	Ex PVinst *pvi;
	Ex SVinst *svi;
	Ex Vtype *vtp;
	Ex Sw nc;  /* Number of cases */
	Ex Sw nv;	/* Number of variables */
	Ex Sw root;
	Ex Class *rootcl;
	Ex AVinst *vlist;
	Ex PVinst *pvars;
	Ex SVinst *svars;

/*	re Classes  */
	Ex Class *cls, *dad;
	Ex Sf cwt;	/*  weight of case in class  */
	Ex Sf scasecost;
	Ex Sw idad;

/*	re Doall   */
	Ex Sw rseed;	/*	Seed for random routines */
	Ex Sw nosubs;
	Ex Sw newsubs;
	Ex Sw deaded;	/* Shows some class killed */
	Ex Sw numson;
	Ex Class *sons[MaxClasses];

/*	re Tuning  */
	Ex Sw MinAge; 	/* Min class age for creation of subs */
	Ex Sw MinSubAge;    /* Until this age, subclass relabs not allowed to
			decay.  */
	Ex Sw MaxSubAge; 	/* Subclasses older than this are replaced  */
	Ex Sw HoldTime;	/* Massage cycles for which class type is fixed
		following forcible change  */
	Ex Sw Forever;		/* If holdtype set to this, it isnt counted down */
	Ex Sf MinSize ;	/* Min size for a class */
	Ex Sw NewSubsTime;	/* Period for making new subclasses */
	Ex Sf MinGain;	/* smallest decrease in cost taken as success */
	Ex Sw GiveUp;		/* Num of nojoy steps until stopping */
	Ex Sw BadMem;		/* Num of cycles a bad move remembered */

/*	re Tactics	*/
	Ex Sw cycle;

/*	re Reports (concerns poplns.c, samples.c)  */
#ifndef NOTGLOB
char *saveheading = "VanillaSnob-Model";
char *trepheading = "VanillaSnob-Thing-Report-File";
char *mrepheading = "VanillaSnob-Member-Report-File";
char *sigstr[] =
  {"INSIG"," weak","   5%","   1%"," 0.1%","0.01%","10-4%"};
Sf siglev [5];
#endif
#ifdef NOTGLOB
extern char *saveheading, *trepheading, *mrepheading;
extern char *sigstr[];
extern Sf siglev[];
#endif

/*	re functions defined in glob.c   */




/*	zzzzzzzzzzzzzzzzzzzzzzzz  */
