#define NOTGLOB 1
#define HTREE 1
#include "glob.c"


/*	----------------------  mmat  ---------------------------   */
/*	Allocates a 'row' by 'col' rectangular matrix and its dope vector.
	Returns ptr to dope vector. Done in such a way that calling 'free()'
	with this ptr will de-allocate the whole thing.
	'isz' is the item size, e.g. 'sizeof(long)' for a matrix of longs.
	*/

void ** mmat(row, col, isz)
	Sw row, col, isz;
{
	Sw ptsz, dvsz, matsz, blsz, i;
	char *block;
	void *mat, **dope;

/*	Get sizes of dope vector, matrix, block  */
	ptsz = sizeof (void *);
	dvsz = row * ptsz;
/*	Round up to a multiple of isz  */
	while (dvsz % isz) dvsz++;
	matsz = row * col * isz;
	blsz = dvsz + matsz;

/*	Allocate a block of size blsz  */
	block = (char*) malloc (blsz);
	dope = (void**) block;
	mat = (void*) (block + dvsz);

/*	Fill in dope vector */
	for (i = 0; i < row; i++)  {
		dope[i] = mat;
		mat += col * isz;
		}
	return (dope);
}


/*	-------------------------  htree  --------------------------  */
/*	Builds a binary tree of classes    */
/*	Uses Battacharyya Coefficient   */
Sw htree ()
{
	Sw nn, ndad, i, j, bi, bj, *use;
	Sw *table, *name, *rhp, *lson, *rson, *centre;
	Sw head, last, line;
	char ** pic;
	Sw lsj, rsj, unsiz;
	Sf **batt, **prb, *size, big;
	Class *clp;

/*	Check we have an attched sample and model  */
	if (! ctx.popln) return (-1);
	if (! ctx.sample) return (-2);
	setpop();
	if (! samp->nc) return (-3);

	findall (Leaf);
	sortbyser();
/*	Make a matrix with a row for each class, a col for each active thing */
	prb = (Sf**) mmat (numson, samp->nact, sizeof(Sf));

	fix = Partial;  control = 0;
	for (nn = 0; nn < samp->nc; nn++)	{
		docase (nn, 0);
		if (! *rec) goto casedone1;
		for (i = 0; i < numson; i++)	{
			clp = sons[i];
			prb[i][nn] = sqrt (clp->casewt);
			}
	casedone1:	;
		}

/*	Make tables for batt coeffs etc.  */
	batt = (Sf**) mmat (numson, numson, sizeof(Sf));
	for (i = 0; i < numson; i++)  {
		for (j = 0; j < i; j++) batt[i][j] = 0.0;
		}

	for (nn = 0; nn < samp->nact; nn++)	{
		for (i = 0; i < numson; i++)  {
			for (j = 0; j < i; j++)  {
				batt[i][j] += prb[i][nn] * prb[j][nn];
				}
			}
		}

	for (i = 0; i < numson; i++)  {
		for (j = 0; j < i; j++)  {
			batt[i][j] = batt[i][j] /
				 	sqrt (sons[i]->cnt * sons[j]->cnt);
			}
		}

/*	Print similarity table  */
	printf ("\n Table of class similarities\n");
	printf (" SER");
	for (i = 0; i < numson-1; i++) printf ("%8d", sons[i]->serial>>2);
	printf ("\n");
	for (i = 1; i < numson; i++)  {
		printf ("%4d", sons[i]->serial>>2);
		for (j = 0; j < i; j++)  {
		    if (batt[i][j] >= 0.001) printf("%8.5f", batt[i][j]);
			else if (batt[i][j] == 0.0) printf(" 0      ");
			else printf ("%8.1e", batt[i][j]);
			}
		printf ("\n");
		}
	printf ("\n");

/*	Now build a tree. When a pair of classes is joined, a row and
	column for one of them in batt[][] is used for the sum, and
	the row and col for the other is killed. To keep track, a vector
	of the user of batt indices is kept.
	*/
	size = (Sf*) malloc (numson * sizeof (Sf));
	use = (Sw*) malloc (numson * sizeof (Sw));
	for (i = 0; i < numson; i++)  {
		use[i] = i;
		size[i] = sons[i]->cnt;
		}

	unsiz = 5;
	table = (Sw*) malloc (2 * numson * unsiz * sizeof (Sw));
	name = table;
	rhp = name + 2*numson;
	lson = rhp + 2*numson;
	rson = lson + 2*numson;
	centre = rson + 2*numson;

/*	The first 'numson' table entries are the leaves  */
	for (i = 0; i < numson; i++)  {
		name[i] = i;
		rhp[i] = 7;
		lson[i] = rson[i] = centre[i] = -1;
		}
	ndad = numson;

joinloop:
/*	Find largest element in batt[][].   */
	big = 0.0;
	for (i = 0; i < numson; i++)  {
		for (j = 0; j < i; j++)  {
			if (batt[i][j] > big)  {
				big = batt[i][j];
				bi = i;  bj = j;
				}
			}
		}
	if (big == 0.0) goto alljoined;

/*	classes at batt indices bi, bj join   */
/*	Note bj < bi, so bj has smaller serial.  Use bj for the join.
	We have to re-compute prb[j][*], batt[j][*], batt[*][j]  */
	printf ("Join ");
	if (use[bj] >= numson) printf (" Dad %3d", use[bj]-numson+1);
	else printf ("Leaf %3d", sons[use[bj]]->serial>>2);
	printf ("  and  ");
	if (use[bi] >= numson) printf (" Dad %3d", use[bi]-numson+1);
	else printf ("Leaf %3d", sons[use[bi]]->serial>>2);
	printf ("  into dad %3d at sim %10.3e\n", ndad-numson+1, big);

/*	Fix links of combining classes, and make table entries for dad. */
	lsj = use[bj];  rsj = use[bi];
	lson[ndad] = lsj;  rson[ndad] = rsj;
/*	Find rightmost leaf in left son, leftmost leaf in right son  */
	while (rson[lsj] >= 0) lsj = rson[lsj];
	while (lson[rsj] >= 0) rsj = lson[rsj];
/*	Join leaf lists  */
	rson[lsj] = rsj;  lson[rsj] = lsj;
	rhp[ndad] = 7 + 2*ndad;

	use[bi] = -1;  use[bj] = ndad;
	ndad++;

/*	If ndad now equal to 2*numson-1, all joins possible are done  */
	if (ndad == (2*numson-1)) goto alljoined;
/*	Zero the affected elements of batt[][]  */
	for (i = 0; i < numson; i++)  {
		if (i > bi) batt[i][bi] = 0.0;
		else if (i < bi) batt[bi][i] = 0.0;
		if (i > bj) batt[i][bj] = 0.0;
		else if (i < bj) batt[bj][i] = 0.0;
		}
	for (nn = 0; nn < samp->nact; nn++)  {
		prb[bj][nn] = sqrt (prb[bj][nn]*prb[bj][nn] +
					prb[bi][nn] * prb[bi][nn]);
		prb[bi][nn] = 0.0;
	/*	Add product to other batt cells  */
		for (i = 0; i < bj; i++)  {
			if (use[i]>=0) batt[bj][i] += prb[bj][nn] * prb[i][nn];
			}
		for (i = bj+1; i < numson; i++) {
			if (use[i]>=0) batt[i][bj] += prb[bj][nn] * prb[i][nn];
			}
		}

	size[bj] += size[bi];  size[bi] = 0.0;
	for (i = 0; i < bj; i++)  {
		if (use[i]>=0) batt[bj][i] = batt[bj][i] /
				sqrt (size[i] * size[bj]);
		}
	for (i = bj+1; i < numson; i++)  {
		if (use[i]>=0) batt[i][bj] = batt[i][bj] /
				sqrt (size[i] * size[bj]);
		}
	goto joinloop;

alljoined:
	printf ("\nBinary tree(s) of classes. There are %2d roots\n",
		2*numson - ndad);

/*	Make a printable image. Needs 2 lines per leaf, (7+2numson) cols. */
	pic = (char**) mmat (2*numson, 1+2*numson, 1);
/*	Clear to spaces  */
	for (i = 0; i < 2*numson; i++)  {
		for (j = 0; j < (2*numson); j++) pic[i][j] = ' ';
		pic[i][2*numson] = 0;
		}

/*	Find all the roots  */
	head = last = -1;
	for (bj = 0; bj < numson; bj++)  {
	    if (use[bj] >= 0)  {  /* It's a root, may be leaf or dad */
	/*	Link into leaf chain  */
		lsj = rsj = use[bj];
		while (rson[lsj] >= 0) lsj = rson[lsj];
		while (lson[rsj] >= 0) rsj = lson[rsj];
	/*	Chain for this root begins at leaf rsj, ends at lsj */
		if (head < 0) head = rsj; else rson[last] = rsj;
		last = lsj;
		}
	    }
	line = 0;
/*	All leaves now linked into chain via rson links. Begin at head */
	for (bi = head; bi >= 0; bi = rson[bi])  {
		centre[bi] = line;
		pic[line][0] = '.';
		for (i = 1; i < 2*numson; i++) pic[line][i] = '_';
		line += 2;
		}
/*	Now do dads  */
	for (bi = numson; bi < ndad; bi++)	{
	/*	Dad at index bi joins two things at indicies lson, rson */
		lsj = centre [lson[bi]];
		rsj = centre [rson[bi]];
		centre [bi] = (lsj + rsj) / 2;
		j = 2 * (bi - numson + 1);
		for (i = lsj+1; i <= rsj; i++) pic [i][j] = '|';
		pic[lsj][j] = '.';
		for (i = j+1; i < 2*numson; i++)  {
			pic [centre[bi]][i] = '_';
			pic [lsj][i] = pic [rsj][i] = ' ';
			}
		}

/*	Write out pic[][]  */
	line = 0;
	for (bi = head; bi >= 0; bi = rson[bi])  {
		printf("%4d %s\n", sons[bi]->serial>>2, pic[line]);
		printf("     %s\n", pic[line+1]);
		line += 2;
		}

	free (pic);
	free (prb);  free (batt);
	free (use);  free (size);
	free (table);
	doall (1);
	fix = dfix;  control = dcontrol;
	return (0);
}


