/*	File of stuff for Multistate variables.   */

#define NOTGLOB 1
#include "glob.c"

#define MaxState 20	/* Max number of discrete states per variable */
static void setvar();
static Sw readvaux();
static Sw readsaux();
static Sw readdat();
static void printdat();
static void setsizes ();
static void clearstats();
static void costvar();
static void derivvar();
static void adjust();
static void vprint();
static Sw inpar();

typedef struct Vauxst {
	Sw states;
	Sf lsprior;	/*  log ((states-1) !) */
	} Vaux;

typedef Sw Datum;

typedef struct Sauxst {
	Sw missing;
	Sf dummy;
	Datum xn;
	} Saux;

typedef struct Pauxst {
	Sw dummy;
	} Paux;

/*	Common variable for holding a data value  */

typedef struct Basicst {     /* Basic parameter info about var in class.
				The first few fields are standard and must
				appear in the order shown.  */
	Sw id;	/* The variable number (index)  */
	Sw signif;
	Sf samplesize;	/* Size of sample on which estimates based */
	/********************  Following fields vary according to need ***/
	Sf origin;	/* This becomes the first element in a number of
			vectors, each of length 'states', which will be
			appended to the Basic sructure. Pointers to these
			vectors will be set in the static pointers declared
			below. A similar thing happens to the Stats structure.
			*/
	}   Basic;

/*	Pointers to vectors in Basic  */
	static Sf *pr;	/* State probs */
	static Sf *lpr; /* State costs inc. roundoff */

/*	Pointer to rootclass lpr  */
	static Sf *dlpr;

typedef struct Statsst  {	/* Stuff accumulated to revise Basic  */
				/* First fields are standard  */
	Sf cnt;		/* Weighted count  */
	Sf stcost, spcost;
	/********************  Following fields vary according to need ***/
	Sf parkstcost;  /* Unweighted thing costs of xn */
	Sf origin;	/* First element of states vectors */
	}  Stats;

/*	Pointers to vectors in stats  */
	static Sf *scnt;	/*  vector of counts in states  */

/*	Static variables useful for many types of variable    */
static Saux *saux;   static Paux *paux;   static Vaux *vaux;
static Basic *cvi, *dcvi;
static Stats *evi;


/*	The macro below abbreviates a scan over states  */
#define Fork for(k=0; k < states; k++)

/*	Static variables specific to this type   */
	static Sw states;
	static Sf statesm;    /*  states - 1.0 */

/*--------------------------  define ------------------------------- */
/*	This routine is used to set up a Vtype entry in the global "types"
array.  It is the only function whose name needs to be altered for different
types of variable, and this name must be copied into the file "definetypes.c"
when installing a new type of variable. It is also necessary to change the
"Ntypes" constant, and to decide on a type id (an integer) for the new type.
	*/

void mults_define(typindx)
	Sw typindx;
/*	typindx is the index in types[] of this type   */
{
	vtp = types + typindx;
	vtp->id = typindx;
/* 	Set type name as string up to 59 chars  */
	vtp->name = "MultiState";
	vtp->datsize = sizeof (Datum);
	vtp->vauxsize = sizeof (Vaux);
	vtp->pauxsize = sizeof (Paux);
	vtp->sauxsize = sizeof (Saux);
	vtp->readvaux = & readvaux;
	vtp->readsaux = & readsaux;
	vtp->readdat = & readdat;
	vtp->printdat = & printdat;
	vtp->setsizes = & setsizes;
	vtp->clearstats = & clearstats;
	vtp->costvar = & costvar;
	vtp->derivvar = & derivvar;
	vtp->adjust = & adjust;
	vtp->vprint = & vprint;
	vtp->inpar = & inpar;
	vtp->setvar = & setvar;

	return;
}


/*	----------------------- setvar --------------------------  */
void setvar (iv)
	Sw iv;
{
	avi = vlist + iv;  vtp = avi->vtp;
	pvi = pvars + iv;  paux = (Paux *) pvi->paux;
	svi = svars + iv;
	vaux = (Vaux *) avi->vaux;
	saux = (Saux *) svi->saux;
	cvi = (Basic*) cls->basics[iv];
	evi = (Stats*) cls->stats[iv];
	dcvi = (Basic*) rootcl->basics[iv];
	states = vaux->states;  statesm = states - 1;
	pr=&(cvi->origin); lpr=pr+states;
	dlpr = &(dcvi->origin) + states;
	scnt = &(evi->origin);
	return;
}


/*	---------------------  readvaux ---------------------------   */
/*	To read any auxiliary info about a variable of this type in some
sample.
	*/
Sw readvaux (vax)
	Vaux *vax;
{
	Sw i;
	Sf lstates;

/*	Read in auxiliary info into vaux, return 0 if OK else 1  */
	i = readint (&(vax->states), 1);
	if (i < 0) {vax->states = 0;  return (1);  }
	if (vax->states > MaxState) {
		printf("Variable has more than %d states (%d)\n",
			MaxState, vax->states);
		return (2);
		}
	lstates = log ((Sf) (vax->states));
	return (0);
}


/*	-------------------  readsaux ------------------------------  */
/*	To read auxilliary info re sample for this attribute   */
Sw readsaux (sax)
	Saux *sax;
{
/*	Multistate has no auxilliary info re sample  */
	return (0);
}


/*	-------------------  readdat -------------------------------  */
/*	To read a value for this variable type         */
Sw readdat (loc, iv)
	char *loc;
	Sw iv;
{
	Sw i;
	Datum xn;

	vaux = (Vaux*) (vlist[iv].vaux);
	states = vaux->states;
/*	Read datum into xn, return error.  */
	i = readint (&xn, 1);
	if (i) return (i);
	if (! xn) return (-1);  /* Missing */
	xn --;
	if ((xn < 0) || (xn >= states)) return (2);
	cmcpy (loc, &xn, sizeof (Datum));
	return (0);
}


/*	---------------------  printdat --------------------------  */
/*	To print a Datum value   */
void printdat (loc)
	Datum *loc;
{
/*	Print datum from address loc   */
	printf ("%3d", (*((Datum*) loc) + 1));
	return;
}

/*	---------------------  setsizes  -----------------------   */
/*	To use info in ctx.vset to set sizes of basic and stats
blocks for variable, and place in AVinst basicsize, statssize.
	*/
void setsizes (iv)
	Sw iv;
{

	avi = vlist + iv;
	vaux = (Vaux*) avi->vaux;  states = vaux->states;

/*	Set sizes of CVinst (basic) and EVinst (stats) in AVinst  */
/*	Each inst has a number of vectors appended, of length 'states' */
	avi->basicsize = sizeof (Basic) + (2*states - 1) * sizeof (Sf);
	avi->statssize = sizeof (Stats) + (states - 1) * sizeof (Sf);
	return;
}

/*	---------------------------  clearstats  --------------------   */
/*	Clears stats to accumulate in costvar, and derives useful functions
of basic params   */
void clearstats (iv)
	Sw iv;
{
	Sw k;

	setvar (iv);
	evi->cnt = 0.0;
	evi->stcost = 0.0;
	Fork scnt[k] = 0.0;
	return;
}


/*	---------------  costvar  ------------------------------  */
/*	Accumulate cost into scasecost  */
void costvar (iv)
	Sw iv;

{
	setvar (iv);
	if (avi->idle) return;
	if (saux->missing) return;
	evi->parkstcost = (cvi->signif) ? lpr[saux->xn] : dlpr[saux->xn];
	scasecost += evi->parkstcost;
	return;
}


/*	------------------  derivvar  ------------------------------  */
/*	      Accumulates weighted statistics  */
void derivvar (iv)
	Sw iv;
{

	setvar (iv);
	if (saux->missing) return;
	evi->cnt += cwt;
	scnt [saux->xn] += cwt;
	evi->stcost += cwt * evi->parkstcost;
	return;
}


/*	-------------------  adjust  ---------------------------    */
/*	To adjust parameters of a multistate variable     */
void adjust (iv)
	Sw iv;
{
	Sf cost, cnt, tcost, dtcost;
	Sw k;

	setvar(iv);
	cnt = evi->cnt;

	if (! (control & AdjPr)) goto adjusted;
	Fork  {
		pr[k] = (scnt[k] + 0.5) / (cnt + states * 0.5);
		}
	cvi->samplesize = cnt;

adjusted:
/*	Compute param costs. Prior dens. uniform = (states-1)!.
	Fisher = cnt^(states-1) * / Prod_k {pr_k}  */
	cost = 0.0;
	Fork  {
		if (k) cost -= log ((Sf) k);
		lpr[k] = - log (pr[k]);
		cost += 0.5 * lpr[k];
		lpr[k] += statesm * 0.5 / cvi->samplesize;
		}
	cost += 0.5 * statesm * log (cvi->samplesize + 1.0);
	evi->spcost = cost;
/*	Get revised thing cost and thing cost using popln params. */
	dtcost = tcost = 0.0;
	Fork {
		tcost += scnt[k] * lpr[k];
		dtcost += scnt[k] * dlpr[k];
		}
/*	If class old enough, test for significance  */
/*	Var is insig if param cost + tcost > dtcost  */
	if (cls->age > MinSubAge)  {
		cvi->signif = setsig (dtcost - evi->spcost - tcost);
		}
	if (cvi->signif) {
		cls->cdtcost += tcost;  cls->cbpcost += evi->spcost;
		}
	else cls->cdtcost += dtcost;
	if ((fix != Random) || (! (control & AdjPr))) return;

/*	Fiddle probs, lpr  */
	cnt = 0.0;
	Fork	{
		scnt[k] += fran() * sqrt (12.0 * scnt[k]) + 0.25;
		cnt += scnt[k];
		}
	cnt = 1.0 / cnt;
	Fork	{
		pr[k] = scnt[k] * cnt;
		lpr[k] = - log (pr[k]) + statesm * 0.5 / cvi->samplesize;
		}
	return;
}

/*	------------------------  vprint  -----------------------   */
void vprint (ccl, iv, fl)
	Class *ccl;
	Sw iv;
	FILE *fl;
{
	Sw k;

	prvhead (ccl, iv, fl);
	setvar (iv);
	fprintf (fl, "State probabilities:\n");
	Fork fprintf (fl, " %2d# %5.3f", k+1, pr[k]);
	fprintf (fl, "\n");
	return;
}


/*	--------------------------  inpar  -------------------  */
/*      To read in the parameters of a multistate distribution   */
Sw inpar (ccl, iv, fl)
        Class *ccl;  Sw iv;  FILE *fl;
{
        Sw err, k;
	Sf sum;

	err = invhead (ccl, iv, fl);  if (err) goto finish;
	setvar (iv);
	sum = 0.0;
	Fork	{
		err = adv (fl);  if (err) goto finish;
		if (fscanf (fl, "%lf", pr+k)  != 1) { err = 4; goto finish; }
		if (pr[k] < 0.0) { err = 24; goto finish; }
		sum += pr[k];
		}
	if (sum > 0.0) sum = 1.0 / sum;  else { err = 25; goto finish; }
	Fork	{
		pr[k] *= sum;
		lpr[k] = - log (pr[k]) + statesm * 0.5 / cvi->samplesize;
		}
finish:
	return (err);
}
/*zzzz*/
