/*	A prototypical file for a variable type      */
/*	actually for plain Gaussians   */

#define NOTGLOB 1
#include "glob.c"

static void setvar();
static Sw readvaux();
static Sw readsaux();
static Sw readdat();
static void printdat();
static void setsizes ();
static void clearstats ();
static void costvar();
static void derivvar();
static void adjust();
static void vprint();
Sw inpar ();

typedef Sf Datum;

typedef struct Sauxst {
	Sw missing;
	Sf dummy;
	Datum xn;
	Sf eps;
	Sf leps;  /* Log eps */
	Sf epssq;	/*  eps^2 / 12  */
	} Saux;

typedef struct Pauxst {
	Sw dummy;
	} Paux;

typedef struct Vauxst {
	Sw dummy;
	} Vaux;

typedef struct Basicst {     /* Basic parameter info about var in class.
				The first few fields are standard and must
				appear in the order shown.  */
	Sw id;	/* The variable number (index)  */
	Sw signif;
	Sf samplesize;  /* Size of sample on which estimates based */
	/********************  Following fields vary according to need ***/
	Sf mu, sig, sdl;  /* Mean, sd, logsd */
		/* sdl = log (sig) + 1.0 / samplesize + hlg2pi - log eps */
	Sf srsds;	/* 1 / sigsq */
	}   Basic;

typedef struct Statsst  {	/* Stuff accumulated to revise Basic  */
				/* First fields are standard  */
	Sf cnt;		/* Weighted count  */
	Sf stcost;
	Sf spcost;
	/********************  Following fields vary according to need ***/
	Sf tx, txx;	/* Sum of vals and squared vals */
	Sf parkstcost;  /* Unweighted thing costs of xn */
	}  Stats;

static Saux *saux;   static Paux *paux;   static Vaux *vaux;
static Basic *cvi, *dcvi;
static Stats *evi;

/*--------------------------  define ------------------------------- */
/*	This routine is used to set up a Vtype entry in the global "types"
array.  It is the only function whose name needs to be altered for different
types of variable, and this name must be copied into the file "definetypes.c"
when installing a new type of variable. It is also necessary to change the
"Ntypes" constant, and to decide on a type id (an integer) for the new type.
	*/

void reals_define(typindx)
	Sw typindx;
/*	typindx is the index in types[] of this type   */
{
	Vtype *vtp;

	vtp = types + typindx;
	vtp->id = typindx;
/* 	Set type name as string up to 59 chars  */
	vtp->name = "Standard Gaussian";
	vtp->datsize = sizeof (Datum);
	vtp->vauxsize = sizeof (Vaux);
	vtp->pauxsize = sizeof (Paux);
	vtp->sauxsize = sizeof (Saux);
	vtp->readvaux = & readvaux;
	vtp->readsaux = & readsaux;
	vtp->readdat = & readdat;
	vtp->printdat = & printdat;
	vtp->setsizes = & setsizes;
	vtp->clearstats = & clearstats;
	vtp->costvar = & costvar;
	vtp->derivvar = & derivvar;
	vtp->adjust = & adjust;
	vtp->vprint = & vprint;
	vtp->inpar = & inpar;
	vtp->setvar = & setvar;
}


/*	-------------------  setvar -----------------------------  */
void setvar (iv)
	Sw iv;
{
	avi = vlist + iv;  vtp = avi->vtp;
	pvi = pvars + iv;  paux = (Paux *) pvi->paux;
	svi = svars + iv;
	vaux = (Vaux *) avi->vaux;
	saux = (Saux *) svi->saux;
	cvi = (Basic*) cls->basics[iv];
	evi = (Stats*) cls->stats[iv];
	dcvi = (Basic*) rootcl->basics[iv];
	return;
}


/*	--------------------  readvaux  ----------------------------  */
/*      Read in auxiliary info into vaux, return 0 if OK else 1  */
Sw readvaux (vax)
	Vaux *vax;
{
	return (0);
}


/*	---------------------  readsaux ---------------------------   */
/*	To read any auxiliary info about a variable of this type in some
sample.
	*/
Sw readsaux (sax)
	Saux *sax;
{
	Sw i;

/*	Read in auxiliary info into saux, return 0 if OK else 1  */
	i = readdf (&(sax->eps), 1);
	if (i < 0) {sax->eps = sax->leps = 0.0;  return (1);  }
	sax->epssq = sax->eps * sax->eps * (1.0/12.0);
	sax->leps = log (sax->eps);   return (0);
}


/*	-------------------  readdat -------------------------------  */
/*	To read a value for this variable type	 */
Sw readdat (loc, iv)
	char *loc;
	Sw iv;
{
	Sw i;
	Datum xn;

/*	Read datum into xn, return error.  */
	i = readdf (&xn, 1);
	if (! i) cmcpy (loc, &xn, sizeof (Datum));
	return (i);
}


/*	---------------------  printdat --------------------------  */
/*	To print a Datum value   */
void printdat (loc)
	char *loc;
{
/*	Print datum from address loc   */
	printf ("%9.2f", * ((Sf*) loc) );
	return;
}

/*	---------------------  setsizes  -----------------------   */
void setsizes (iv)
	Sw iv;
{
	avi = vlist + iv;
	avi->basicsize = sizeof (Basic);
	avi->statssize = sizeof (Stats);
	return;
}

/*	------------------------  clearstats  ------------------------  */
/*	Clears stats to accumulate in costvar, and derives useful functions
of basic params  */
void clearstats (iv)
	Sw iv;
{
	setvar (iv);
	evi->cnt = 0.0;
	evi->stcost = 0.0;
	evi->tx = evi->txx = 0.0;

	return;
}


/*	-----------------------  costvar  --------------------------   */
/*	Accumulates thing cost into scasecost, fcasecost    */
void costvar (iv)
	Sw iv;
{
	Sf del, var, cost;
	setvar (iv);
	if (saux->missing) return;
	if (! cvi->signif) cvi = dcvi;
	del = cvi->mu - saux->xn;
	var = del * del + saux->epssq;
	cost = 0.5 * var * cvi->srsds + cvi->sdl;
	evi->parkstcost = cost;
	scasecost += cost;

	return;
}


/*	--------------------  derivvar  --------------------------  */
/*	Accumulates weighted statistics  */
void derivvar (iv)
	Sw iv;
{
	setvar (iv);
	if (saux->missing) return;
	evi->cnt += cwt;
	evi->tx += cwt * saux->xn;
	evi->txx += cwt * (saux->xn * saux->xn + saux->epssq);
/*	Accumulate weighted thing cost  */
	evi->stcost += cwt * evi->parkstcost;
	return;
}


/*	-------------------  adjust  ---------------------------    */
/*	To adjust parameters of a real variable     */
void adjust (iv)
	Sw iv;
{
	Sf cnt, var, vari, mean, tcost, dtcost;

	setvar (iv);
	cnt = evi->cnt;
	if (cnt < 1.0) {
		mean = cvi->mu;  var = cvi->sig*cvi->sig;
		goto adjusted;
		}
	mean = evi->tx / cnt;
	var = evi->txx - cnt * mean * mean;

	if (! (control & AdjPr)) goto adjusted;
	cvi->mu = mean;
	if (cnt >= 2.0) cvi->sig = sqrt (var / (cnt - 1.0));
	if (fix == Random)  {
		vari = 1.0 / sqrt (cnt + 4.0);
		cvi->mu += 3.464 * (fran() - 0.5) * vari * cvi->sig;
		cvi->sig *= 1.0 + 2.449 * (fran() - 0.5) * vari;
		}
	cvi->srsds = 1.0 / (cvi->sig * cvi->sig);
	cvi->samplesize = cnt;
	cvi->sdl = log (cvi->sig) + 1.0 / cvi->samplesize
			+ hlg2pi - saux->leps;

adjusted:
/*	Compute parameter costs. Prior dens. of mu is uniform in range of
size 6 * rootcl->sd, prior dens. of sig is uniform in range rootcl->sd.
Fisher for mu is cnt/sigsq, for sig is 2*cnt/sigsq  */
/*	The ranges 6sd and 2sd nicely cancel the 2 root(12)s for the params */

	evi->spcost = log (dcvi->sig * dcvi->sig *
			cvi->srsds * cvi->samplesize);

/*	Get revised tcost and tcost using popln params  */
/*	Note that if adjustment has not been done, variance of data
	about mu is not yet known. Let us recompute it anyway, it does not
	take long.  */
	vari = var + cnt * (cvi->mu - mean) * (cvi->mu - mean);
	tcost = 0.5 * cvi->srsds * vari + cnt * cvi->sdl;
	vari = var + cnt * (dcvi->mu - mean) * (dcvi->mu - mean);
	dtcost = 0.5 * dcvi->srsds * vari + cnt * dcvi->sdl;
/*	If class old enough, test for significance  */
/*	Var is insig if param cost + tcost > dtcost  */
        if (cls->age > MinSubAge)  {
                cvi->signif = setsig (dtcost - evi->spcost - tcost);
                }
        if (cvi->signif) {
                cls->cdtcost += tcost;  cls->cbpcost += evi->spcost;
                }
        else cls->cdtcost += dtcost;
	return;
}


/*	------------------------  vprint  -----------------------   */
void vprint (ccl, iv, fl)
	Class *ccl;
	Sw iv;
	FILE *fl;
{

	prvhead (ccl, iv, fl);
	setvar (iv);
	fprintf (fl, "  Mu# %15.6f  SD# %14.6f\n", cvi->mu, cvi->sig);
	return;
}


/*	-----------------------  inpar  --------------------------  */
/*	To read in the parameters of a real distribution   */
Sw inpar (ccl, iv, fl)
	Class *ccl;  Sw iv;  FILE *fl;
{
	Sw err, j;
	Sf pars[2];

	err = invhead (ccl, iv, fl);  if (err) goto finish;
	setvar (iv);
	for (j = 0; j < 2; j++)	{
		err = adv (fl);  if (err) goto finish;
		if (fscanf (fl, "%lf", pars+j) != 1) { err = 4; goto finish; }
		}
	cvi->mu = pars[0];  cvi->sig = pars[1];
	if (cvi->sig <= 0.0) { err = 5;  goto finish; }
	cvi->sdl = log (cvi->sig) + 1.0 / cvi->samplesize
			+ hlg2pi - saux->leps;
	cvi->srsds = 1.0 / (cvi->sig * cvi->sig);
	err = 0;
finish:
	return (err);
}
